/************************************************************************/
/*  seg2toms -	convert a SEG2 file to MiniSEED.			*/
/*									*/
/*	Cedric de La Beaujardiere 					*/
/*	Seismological Laboratory					*/
/*	University of California, Berkeley				*/
/*									*/
/*	Original segy2ms.c written by:    				*/
/*	Douglas Neuhauser						*/
/*	doug@seismo.berkeley.edu					*/
/*									*/
/************************************************************************/

/*
 * Copyright (c) 1996-2004 The Regents of the University of California.
 * All Rights Reserved.
 * 
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for educational, research and non-profit purposes,
 * without fee, and without a written agreement is hereby granted,
 * provided that the above copyright notice, this paragraph and the
 * following three paragraphs appear in all copies.
 * 
 * Permission to incorporate this software into commercial products may
 * be obtained from the Office of Technology Licensing, 2150 Shattuck
 * Avenue, Suite 510, Berkeley, CA  94704.
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY
 * FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES,
 * INCLUDING LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE AND
 * ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF CALIFORNIA HAS BEEN
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE
 * PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND THE UNIVERSITY OF
 * CALIFORNIA HAS NO OBLIGATIONS TO PROVIDE MAINTENANCE, SUPPORT,
 * UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 */

#ifndef lint
static char sccsid[] = "$Id: seg2toms.c,v 1.9 2006/10/11 22:42:30 doug Exp $ ";
#endif

#include <stdio.h>
#include <string.h>
#include <malloc.h>
#include <stdlib.h>

#include "version.h"

#include "qlib2.h"

#include "seg2_struct.h"
#include "procs.h"

#ifndef	DEFAULT_NETWORK
#define	DEFAULT_NETWORK		"SF"
#endif

#ifndef	DEFAULT_OUTPUT_FORMAT
#define	DEFAULT_OUTPUT_FORMAT	STEIM2
#endif

#ifndef	DEFAULT_BLKSIZE
#define	DEFAULT_BLKSIZE		4096
#endif

#define	EXTERN
#define	DEFINE

#include "externals.h"

char *syntax[] = {
"%s version " VERSION ,
"%s  [-h] [-s] [-S station] [-N network] [-C channel] [-L location]",
"    [-n] [-o outfile] [-m channel_map_file] [-x scale_factor]",
"    [-b blksize] [-O format] [-D date] [-F key=val] [-T key=val] [-d dbgflag]",
"    [input file]",
"    where:",
"	-h	    Help - prints syntax message.",
"	-s	    Output only a summary line with:",
"			date, time, number of traces, and sample rate (Hz)",
"	-S station  Specify explicit station name.",
"	-N network  Specify explicit network name.",
"	-C channel  Specify explicit channel name.",
"	-L location Specify explicit location name.",
"		    Default network name is " DEFAULT_NETWORK ".",
"	-n	    No data output.  Write debug to stdout.",
"	-o output_file",
"		    Specifies an explicit output filename for the MiniSEED data.",
"		    If no output file is specified, write data to stdout.",
"	-m channel_map_file",
"		    Specifies a file that contains information for mapping SEG2",
"		    channel number to SEED SNCL.",
"	-x scale_factor",
"		    Scaling factor for converting floating point data to integer.",
"		    Default value is " STRING(DFLT_FLOAT_SCALING_FACTOR) ".",
"	-b blksize  Specifies output blksize.  Default blocksize=" STRING(DEFAULT_BLKSIZE) ".",
"	-O format   Specifies output format for MiniSEED data.",
"		    Default output format is " STRING(DEFAULT_OUTPUT_FORMAT) ".",
"	-D date     Specify explicit Acquisition Date.",
"		    Default is obtained from SEG2 file descriptor strings section",
"		    ACQUISITION_DATE.  Format is DD/MM/YYYY",
"	-d dbgflag  Prints debug information, according to flag bits set:",
"			  1 = file descriptor number of traces and trace ptrs.",
"			  2 = file descriptor values.",
"			  4 = file descriptor strings (keywd=value).",
"			  8 = trace descriptor values.",
"			 16 = trace descriptors strings (keywd=value).",
"			 32 = trace data min,max data values.",
"			 64 = trace data raw data  values.",
"	-F key=val  Override or add the specified file_descriptor key value string.",
"	-T key=val  Override or add the specified trace_descriptor key value string.",
"	input_files An optional input file containing SEG2 data.",
"		    If no input file is specified, data is read from stdin.",
"Notes:",
"1.  Channel_map_file has the following format:",
"    Comment lines begin with '#'.",
"    Mapping lines have 4 tokens:",
"	tracenumber station network channel location",
"    eg:",
"	1 PH001 GP1 01",
"    If SEED station, channel, and location are all '-', skip this channel.",
"    If any of the SEED fields are '*', use default mapping for that value.",
"    If ALL of the SEED fields are '-', skip this channel.",
"    Default mapping is for each set of 3 input channels to map to SEED channels:",
"	GP1, GP2, GP3 with the same location code values (starting at 01).",
"    Location code will increment after each set of 3 input channels.",
"Examples:",
"1.  Convert a SEG2 file to miniSEED.",
"	%s 2283.dat",
"2.  Provide summary of date, time, #traces, and sample rate.",
"	%s -s 2283.dat",
"3.  Override or add a bad date and time when converting a SEG2 file to MiniSEED.",
"	%s -F ACQUISITION_DATE=05/11/2004 -F ACQUISITION_TIME=01:23:45 2283.dat",
NULL };

/************************************************************************/
/*  main:   main program.						*/
/************************************************************************/
main (int argc, char **argv)
{
  int            i, ival, npts;
  EXT_TIME       et;
  DATA_HDR      *hdr   = NULL;
  int           *data  = NULL;
  char          *mseed = NULL;		/* allow qlib to allocate buffer.	*/
  char          *p;
  int            n, nleft;
  int            status;
  int            nblocks;
  SEG2FILEDESC  *file_desc  = NULL;
  SEG2TRACEDESC *trace_desc = NULL;
  int            ntraces;
  char           filename[2048];

  info = stdout;
  get_my_wordorder();
  set_hdr_wordorder(SEED_BIG_ENDIAN);
  set_data_wordorder(SEED_BIG_ENDIAN);
  init_file_and_trace_desc_keywords ();
  parse_cmdline(&argc,&argv);

  do {
    if (argc > 0) {
      if ((input=fopen(*argv,"r")) == NULL) {
        fprintf (stderr, "Error opening file %s\n", *argv);
        return(1);
      }
    }
    else input = stdin;

    /*  Allocate space for the File Descriptor  */
    if (file_desc != NULL)
      free_file_desc(&file_desc);
    file_desc = (SEG2FILEDESC *)malloc(sizeof(SEG2FILEDESC));
    if (file_desc == NULL) {
      fprintf (stderr,
               "Error in seg2toms main on input %s: Can't allocate memory block for file_desc.\n",
               *input);
      return(1);
    }
    memset (file_desc, 0, sizeof(SEG2FILEDESC));

    status = read_seg2_file_desc (input, file_desc, &ntraces);
    if (status != 0) {
      fprintf (stderr, "Error reading SEG2 file descriptor block in input file %s. Skipping SEG2 file.\n", *argv);
      continue;
/*    return(1);*/
    }

    /* Check for command-line overrides of file descriptor strings. */
    if (orfv) override_file_key_val (file_desc, orfv, orfc);

    for (i=0; i<ntraces; i++) {
      /*  Allocate space for the Trace Descriptor  */
      if (trace_desc != NULL) free_trace_desc(&trace_desc);
      trace_desc = (SEG2TRACEDESC *)malloc(sizeof(SEG2TRACEDESC));
      if (trace_desc == NULL) {
        fprintf(stderr,
                "Error in seg2toms main on input %s: Can't allocate memory block for trace_desc. Skipping trace.\n",
                *input);
        continue;
      }
      memset (trace_desc, 0, sizeof(SEG2TRACEDESC));

      /* Seek to the start of the trace, using its pointer in the file descriptor */
      status = seek_to_trace(input, file_desc->trace_ptrs[i]);
      if (status < 0) {
        fprintf(stderr, "Error: seek_to_trace failed. Skipping trace.\n");
        free_trace_desc(&trace_desc);
        continue;
      }
      
      /*  Read the Trace Descriptor */
      status = read_seg2_trace_desc(input, file_desc, trace_desc);
      if (status != 0) {
        fprintf (stderr, "error reading trace desc block %d in input file %s. Skipping SEG2 trace.\n", i, *argv);
        free_trace_desc(&trace_desc);
        continue;
/*      return(1);*/
      }

      /* Check for command-line overrides of trace descriptor strings. */
      if (ortv) override_trace_key_val (trace_desc, ortv, ortc);

      /* Output summary if requested. */
      if (summary_flag) {
	printf ("date=%s time=%s ntraces=%d rate=%.0lf\n",
		(file_desc->known_keys)[ACQUISITION_DATE] && (file_desc->known_keys)[ACQUISITION_DATE]->val ?
		(file_desc->known_keys)[ACQUISITION_DATE]->val : "(Null)",
		(file_desc->known_keys)[ACQUISITION_TIME] && (file_desc->known_keys)[ACQUISITION_TIME]->val ?
		(file_desc->known_keys)[ACQUISITION_TIME]->val : "(Null)",
		file_desc->numOfTraces,
		(trace_desc->known_keys)[SAMPLE_INTERVAL] && (trace_desc->known_keys)[SAMPLE_INTERVAL]->val ?
		(1.0/atof((trace_desc->known_keys)[SAMPLE_INTERVAL]->val)) : 0.0);
	exit(0);
      }

      /* Check if the size of the trace descriptor is divisible by 4, because *
       * all file and trace descriptor blocks, their strings subblocks, and   *
       * the data blocks must begin on double word boundaries.                *
       * If not, move the file pointer until it is at the next boundary.      */
      if((trace_desc->traceDescBlockSize % 4) != 0) {
        if (fseek(input, (4 - (trace_desc->traceDescBlockSize % 4)), SEEK_CUR)) {
          fprintf(stderr,
                  "Error: while reading SEG2 File %s, seek forward %d bytes to reach double word boundary failed.\n",
                  *input, (4 - (trace_desc->traceDescBlockSize % 4)));
          free_trace_desc(&trace_desc);
          continue;
        }
      }

      /*   Read the Data   */
      status = read_seg2_trace_data(input, file_desc, trace_desc, &hdr, &data);
      if (status == 4) {
        /* Don't process the trace of this seg-2 channel. */
        free_trace_desc(&trace_desc);
        free_data_hdr  (hdr);
        free           (data);
	hdr  = NULL;
	data = NULL;
        continue;
      }
      else if (status != 0 || hdr == NULL || data == NULL) {
        fprintf (stderr, "error reading data block %d in input file %s. Skipping SEG2 trace.\n", i, *argv);
        free_trace_desc(&trace_desc);
        continue;
/*      return(1);*/
      }

      /* Discard traces with no station name.			*/
      if (strlen(hdr->station_id) == 0) continue;

      hdr->blksize = blksize;
      hdr->data_type = output_format;
      npts = hdr->num_samples;
      mseed = NULL;
      status = ms_pack_data (hdr, NULL, npts, data, &nblocks, &mseed, 0, NULL);
      if (status != npts) {
        fprintf(stderr, "ms_pack_data returned error status %d\n", status);
        fflush(NULL); /* output all buffered output data, including message above */
        FATAL("Error packing data into MiniSEED.");
      }
  
      /* Write MiniSEED data to output file.			*/
      et = int_to_ext(hdr->begtime);
      /*
      sprintf (filename, "%s.%s.%s.D.%04d.%03d.%02d%02d%02d",
  	     hdr->station_id, hdr->network_id, hdr->channel_id,
  	     et.year, et.doy, et.hour, et.minute, et.second);
      if ((output = fopen (filename, "w")) == NULL)
  	FATAL("Error opening output file\n");
      */
      n = nleft = nblocks;
      p = (char *)mseed;
      while (nleft > 0 && n > 0) {
  	n = fwrite (p, blksize, nleft, output);
  	p += n*blksize;
  	nleft -= n;
      }
      if (nleft) {
        fprintf (info, "Error writing MiniSEED data - ");
        fprintf (info, "%d records unwritten\n", nleft);
        fprintf (stderr, "error writing MiniSEED data for trace %d in input file %s. %d records unwritten.\n", i, *argv, nleft);
        /* return(1); // Don't abort, just continue to next trace. */
      }
      free_trace_desc(&trace_desc);
      free_data_hdr  (hdr);
      free           (data);
      free           (mseed);
      hdr   = NULL;
      data  = NULL;
      mseed = NULL;
    }
    free_file_desc(&file_desc);
    fclose (input);
    --argc;
    ++argv;
  } while (argc > 0);

  free_keywords();
  fclose (output);
  return(0);
}


char init_file_and_trace_desc_keywords ( void ) {
  char returnval = 0;

  /*** Allocate space for and assign values to file_desc_keywords ***/
  file_desc_keywords =   (SEG2_KEYWORDS *)malloc(sizeof(SEG2_KEYWORDS));
  if (file_desc_keywords == NULL) {
    fprintf (stderr, "seg2toms error: init of keywords unable to malloc space for file_desc_keywords.\n");
    return (1);
  }

  file_desc_keywords->keywords = (char **)malloc(sizeof(char *) * NUM_OF_FD_KEYWORDS);
  if (file_desc_keywords->keywords == NULL) {
    fprintf (stderr,
             "seg2toms error: init of keywords unable to malloc space for file_desc_keywords->keywords, char * array of %d ptrs.\n",
             NUM_OF_FD_KEYWORDS);
    return (1);
  }

  file_desc_keywords->current_index = 0;
  file_desc_keywords->max_index = MAX_FD_KEY_INDEX;
  file_desc_keywords->keywords[ACQUISITION_DATE]            = "ACQUISITION_DATE";
  file_desc_keywords->keywords[ACQUISITION_SECOND_FRACTION] = "ACQUISITION_SECOND_FRACTION";
  file_desc_keywords->keywords[ACQUISITION_TIME]            = "ACQUISITION_TIME";
  file_desc_keywords->keywords[CLIENT]                      = "CLIENT";
  file_desc_keywords->keywords[COMPANY]                     = "COMPANY";
  file_desc_keywords->keywords[GENERAL_CONSTANT]            = "GENERAL_CONSTANT";
  file_desc_keywords->keywords[INSTRUMENT]                  = "INSTRUMENT";
  file_desc_keywords->keywords[JOB_ID]                      = "JOB_ID";
  file_desc_keywords->keywords[OBSERVER]                    = "OBSERVER";
  file_desc_keywords->keywords[PROCESSING_DATE]             = "PROCESSING_DATE";
  file_desc_keywords->keywords[PROCESSING_TIME]             = "PROCESSING_TIME";
  file_desc_keywords->keywords[TRACE_SORT]                  = "TRACE_SORT";
  file_desc_keywords->keywords[UNITS]                       = "UNITS";
  file_desc_keywords->keywords[FD_NOTE]                     = "NOTE";

  /*** Allocate space for and assign values to trace_desc_keywords ***/
  trace_desc_keywords =   (SEG2_KEYWORDS *)malloc(sizeof(SEG2_KEYWORDS));
  if (trace_desc_keywords == NULL) {
    fprintf (stderr, "seg2toms error: init of keywords unable to malloc space for trace_desc_keywords.\n");
    return (1);
  }

  trace_desc_keywords->keywords = (char **)malloc(sizeof(char *) * NUM_OF_TD_KEYWORDS);
  if (trace_desc_keywords->keywords == NULL) {
    fprintf (stderr,
             "seg2toms error: init of keywords unable to malloc space for trace_desc_keywords->keywords, char * array of %d ptrs.\n",
             NUM_OF_TD_KEYWORDS);
    return (1);
  }

  trace_desc_keywords->current_index = 0;
  trace_desc_keywords->max_index = MAX_TD_KEY_INDEX;
  trace_desc_keywords->keywords[ALIAS_FILTER]               = "ALIAS_FILTER";
  trace_desc_keywords->keywords[AMPLITUDE_RECOVERY]         = "AMPLITUDE_RECOVERY";
  trace_desc_keywords->keywords[BAND_REJECT_FILTER]         = "BAND_REJECT_FILTER";
  trace_desc_keywords->keywords[CDP_NUMBER]                 = "CDP_NUMBER";
  trace_desc_keywords->keywords[CDP_TRACE]                  = "CDP_TRACE";
  trace_desc_keywords->keywords[CHANNEL_NUMBER]             = "CHANNEL_NUMBER";
  trace_desc_keywords->keywords[DATUM]                      = "DATUM";
  trace_desc_keywords->keywords[DELAY]                      = "DELAY";
  trace_desc_keywords->keywords[DESCALING_FACTOR]           = "DESCALING_FACTOR";
  trace_desc_keywords->keywords[DIGITAL_BAND_REJECT_FILTER] = "DIGITAL_BAND_REJECT_FILTER";
  trace_desc_keywords->keywords[DIGITAL_LOW_CUT_FILTER]     = "DIGITAL_LOW_CUT_FILTER";
  trace_desc_keywords->keywords[END_OF_GROUP]               = "END_OF_GROUP";
  trace_desc_keywords->keywords[FIXED_GAIN]                 = "FIXED_GAIN";
  trace_desc_keywords->keywords[HIGH_CUT_FILTER]            = "HIGH_CUT_FILTE";
  trace_desc_keywords->keywords[LINE_ID]                    = "LINE_ID";
  trace_desc_keywords->keywords[LOW_CUT_FILTER]             = "LOW_CUT_FILTER";
  trace_desc_keywords->keywords[NOTCH_FREQUENCY]            = "NOTCH_FREQUENCY";
  trace_desc_keywords->keywords[POLARITY]                   = "POLARITY";
  trace_desc_keywords->keywords[RAW_RECORD]                 = "RAW_RECORD";
  trace_desc_keywords->keywords[RECEIVER]                   = "RECEIVER";
  trace_desc_keywords->keywords[RECEIVER_GEOMETRY]          = "RECEIVER_GEOMETRY";
  trace_desc_keywords->keywords[RECEIVER_LOCATION]          = "RECEIVER_LOCATION";
  trace_desc_keywords->keywords[RECEIVER_SPECS]             = "RECEIVER_SPECS";
  trace_desc_keywords->keywords[RECEIVER_STATION_NUMBER]    = "RECEIVER_STATION_NUMBER";
  trace_desc_keywords->keywords[SAMPLE_INTERVAL]            = "SAMPLE_INTERVAL";
  trace_desc_keywords->keywords[SHOT_SEQUENCE_NUMBER]       = "SHOT_SEQUENCE_NUMBER";
  trace_desc_keywords->keywords[SKEW]                       = "SKEW";
  trace_desc_keywords->keywords[SOURCE]                     = "SOURCE";
  trace_desc_keywords->keywords[SOURCE_LOCATION]            = "SOURCE_LOCATION";
  trace_desc_keywords->keywords[SOURCE_STATION_NUMBER]      = "SOURCE_STATION_NUMBER";
  trace_desc_keywords->keywords[STACK]                      = "STACK";
  trace_desc_keywords->keywords[STATIC_CORRECTIONS]         = "STATIC_CORRECTIONS";
  trace_desc_keywords->keywords[TRACE_TYPE]                 = "TRACE_TYPE";
  trace_desc_keywords->keywords[TD_NOTE]                    = "NOTE";

  return (returnval);
}

int seek_to_trace(FILE	*input, int new_position) {
  int   status = 0;
  if (fseeko (input, new_position, SEEK_SET) != 0) {
    fprintf(stderr, "Error: Seeking to file offset.\n");
    status = -1;
    return (status);
  }
  return (status);
}

void free_trace_desc(SEG2TRACEDESC **trace_desc) {
  if (*trace_desc) {
    if ((*trace_desc)->strings) {
      free               ((*trace_desc)->strings);
    }
    if ((*trace_desc)->known_keys) {
      free_key_val_pairs ((*trace_desc)->known_keys);
    }
    if ((*trace_desc)->unknown_keys) {
      free_key_val_pairs ((*trace_desc)->unknown_keys);
      free               ((*trace_desc)->unknown_keys);
    }
    free               (*trace_desc);
    *trace_desc = NULL;
  }
}

void free_key_val_pairs (KEY_VAL_PAIR **pairs) {
  int length = sizeof pairs / sizeof pairs[0];

  while (length > 0) {
    if (pairs[length - 1] != NULL) {
      free (pairs[length - 1]->key);
      free (pairs[length - 1]->val);
    }
    length--;
  }
}

void free_keywords(void) {
  if (file_desc_keywords) {
    if (file_desc_keywords->keywords) {
      free(file_desc_keywords->keywords);
    }
    free(file_desc_keywords);
    file_desc_keywords  = NULL;
  }
  if (trace_desc_keywords) {
    if (trace_desc_keywords->keywords) {
      free(trace_desc_keywords->keywords);
    }
    free(trace_desc_keywords);
    trace_desc_keywords = NULL;
  }
}

void free_file_desc(SEG2FILEDESC **file_desc) {
  if (*file_desc) {
    if ((*file_desc)->strings) {
      free               ((*file_desc)->strings);
    }
    if ((*file_desc)->trace_ptrs) {
      free               ((*file_desc)->trace_ptrs);
    }
    if ((*file_desc)->known_keys) {
      free_key_val_pairs ((*file_desc)->known_keys);
    }
    if ((*file_desc)->unknown_keys) {
      free_key_val_pairs ((*file_desc)->unknown_keys);
      free               ((*file_desc)->unknown_keys);
    }
    free               (*file_desc);
    *file_desc = NULL;
  }
}

